@@ -18,10 +18,10 @@
 
 use self::errors::*;
 use crate::lalr::TokenType;
+use dmntk_common::Result;
 use dmntk_feel::context::FeelContext;
-use dmntk_feel::dmntk_common::Result;
-use dmntk_feel::values::VALUE_NULL;
-use dmntk_feel::{Name, Scope};
+use dmntk_feel::values::Value;
+use dmntk_feel::{value_null, Name, Scope};
 
 /// Definition of a single space character.
 const WS: char = ' ';
@@ -37,6 +37,7 @@ const DECIMAL_SEPARATOR: char = '.';
 pub enum TokenValue {
   YyEmpty,
   YyEof,
+  YyError,
   YyUndef,
   YyState(usize),
   And,
@@ -44,6 +45,7 @@ pub enum TokenValue {
   Between,
   BetweenAnd,
   Boolean(bool),
+  BuiltInTypeName(Name),
   Colon,
   Comma,
   Context,
@@ -87,6 +89,7 @@ pub enum TokenValue {
   RightParen,
   Satisfies,
   Some,
+  StartExpression,
   StartBoxedExpression,
   StartTextualExpression,
   StartTextualExpressions,
@@ -115,10 +118,12 @@ pub struct Lexer<'lexer> {
   /// When this flag is set, the next `and` token is returned as `band` keyword,
   /// otherwise it is returned as `and`. This allows to disambiguate the `and`
   /// operator used in between clause from conjunction.
-  /// After consuming first `and` ad `band` this flag is reset.
-  between_begin: bool,
-  ///
+  /// After consuming first `and` as `band` this flag is cleared by the lexer.
+  between: bool,
+  /// ???
   type_name: bool,
+  /// ???
+  till_in: bool,
 }
 
 /// FEEL lexer implementation.
@@ -131,27 +136,28 @@ impl<'lexer> Lexer<'lexer> {
       input: input.chars().collect(),
       position: 0,
       unary_tests: false,
-      between_begin: false,
+      between: false,
       type_name: false,
+      till_in: false,
     }
   }
 
-  pub fn scope(&self) -> &Scope {
-    &self.scope
-  }
-
   pub fn set_unary_tests(&mut self) {
     self.unary_tests = true;
   }
 
-  pub fn set_between_begin(&mut self) {
-    self.between_begin = true;
+  pub fn set_between(&mut self) {
+    self.between = true;
   }
 
   pub fn set_type_name(&mut self) {
     self.type_name = true;
   }
 
+  pub fn set_till_in(&mut self) {
+    self.till_in = true;
+  }
+
   pub fn push_to_scope(&mut self) {
     self.scope.push(FeelContext::default());
   }
@@ -161,21 +167,22 @@ impl<'lexer> Lexer<'lexer> {
   }
 
   pub fn add_name_to_scope(&mut self, name: &Name) {
-    self.scope.set_entry(name, VALUE_NULL);
+    self.scope.set_entry(name, value_null!());
   }
 
-  ///
-  pub fn get_next_token(&mut self) -> Result<(TokenType, TokenValue)> {
+  /// Returns the next token from input.
+  pub fn next_token(&mut self) -> Result<(TokenType, TokenValue)> {
     if let Some(start_token_type) = self.start_token_type.clone() {
       self.start_token_type = None;
-      match start_token_type {
-        tt @ TokenType::StartBoxedExpression => return Ok((tt, TokenValue::StartBoxedExpression)),
-        tt @ TokenType::StartContext => return Ok((tt, TokenValue::StartBoxedExpression)),
-        tt @ TokenType::StartTextualExpression => return Ok((tt, TokenValue::StartTextualExpression)),
-        tt @ TokenType::StartTextualExpressions => return Ok((tt, TokenValue::StartTextualExpressions)),
-        tt @ TokenType::StartUnaryTests => return Ok((tt, TokenValue::StartUnaryTests)),
-        _ => {}
-      }
+      return Ok(match start_token_type {
+        tt @ TokenType::StartExpression => (tt, TokenValue::StartExpression),
+        tt @ TokenType::StartBoxedExpression => (tt, TokenValue::StartBoxedExpression),
+        tt @ TokenType::StartContext => (tt, TokenValue::StartBoxedExpression),
+        tt @ TokenType::StartTextualExpression => (tt, TokenValue::StartTextualExpression),
+        tt @ TokenType::StartTextualExpressions => (tt, TokenValue::StartTextualExpressions),
+        tt @ TokenType::StartUnaryTests => (tt, TokenValue::StartUnaryTests),
+        _ => (TokenType::YyError, TokenValue::YyError),
+      });
     }
     let result = self.read_next_token();
     self.unary_tests = false;
@@ -194,7 +201,7 @@ impl<'lexer> Lexer<'lexer> {
         self.position += 8;
         Ok((TokenType::External, TokenValue::External))
       }
-      ['f', 'u', 'n', 'c', 't', 'i', 'o', 'n', ch, _, _, _] if is_function_separator(ch) => {
+      ['f', 'u', 'n', 'c', 't', 'i', 'o', 'n', _, _, _, _] if self.is_function_separator(8) => {
         self.position += 8;
         Ok((TokenType::Function, TokenValue::Function))
       }
@@ -206,7 +213,7 @@ impl<'lexer> Lexer<'lexer> {
         self.position += 7;
         Ok((TokenType::Between, TokenValue::Between))
       }
-      ['c', 'o', 'n', 't', 'e', 'x', 't', ch, _, _, _, _] if is_context_separator(ch) => {
+      ['c', 'o', 'n', 't', 'e', 'x', 't', _h, _, _, _, _] if self.is_context_separator(7) => {
         self.position += 7;
         Ok((TokenType::Context, TokenValue::Context))
       }
@@ -222,7 +229,7 @@ impl<'lexer> Lexer<'lexer> {
         self.position += 5;
         Ok((TokenType::Boolean, TokenValue::Boolean(false)))
       }
-      ['r', 'a', 'n', 'g', 'e', ch, _, _, _, _, _, _] if is_range_separator(ch) => {
+      ['r', 'a', 'n', 'g', 'e', _h, _, _, _, _, _, _] if self.is_range_separator(5) => {
         self.position += 5;
         Ok((TokenType::Range, TokenValue::Range))
       }
@@ -234,7 +241,7 @@ impl<'lexer> Lexer<'lexer> {
         self.position += 4;
         Ok((TokenType::Else, TokenValue::Else))
       }
-      ['l', 'i', 's', 't', ch, _, _, _, _, _, _, _] if is_separator(ch) => {
+      ['l', 'i', 's', 't', _, _, _, _, _, _, _, _] if self.is_list_separator(4) => {
         self.position += 4;
         Ok((TokenType::List, TokenValue::List))
       }
@@ -250,12 +257,12 @@ impl<'lexer> Lexer<'lexer> {
         self.position += 4;
         Ok((TokenType::Boolean, TokenValue::Boolean(true)))
       }
-      ['a', 'n', 'd', WS, _, _, _, _, _, _, _, _] if !self.between_begin => {
+      ['a', 'n', 'd', WS, _, _, _, _, _, _, _, _] if !self.between => {
         self.position += 3;
         Ok((TokenType::And, TokenValue::And))
       }
-      ['a', 'n', 'd', WS, _, _, _, _, _, _, _, _] if self.between_begin => {
-        self.between_begin = false;
+      ['a', 'n', 'd', WS, _, _, _, _, _, _, _, _] if self.between => {
+        self.between = false;
         self.position += 3;
         Ok((TokenType::BetweenAnd, TokenValue::BetweenAnd))
       }
@@ -597,8 +604,8 @@ impl<'lexer> Lexer<'lexer> {
         _ => {}
       }
     }
-    // now the `parts` contains all parts of the longest possible name,
-    // now decide what kind of name it is, by checking the parsing scope
+    // now the `parts` vector contains all parts of the longest possible name,
+    // now must be decides what kind of name it is, by checking the parsing scope
 
     // ------------------------------------------------------------------------
     // tweak with name of the `item` in filter
@@ -610,6 +617,20 @@ impl<'lexer> Lexer<'lexer> {
       }
     }
 
+    // ------------------------------------------------------------------------
+    // tweak with the name in `for` and `quantified` expressions
+    // variable name is the name before the keyword `in`
+    // ------------------------------------------------------------------------
+    if self.till_in {
+      if let Some(index) = parts.iter().position(|value| value == "in") {
+        self.till_in = false;
+        parts.truncate(index);
+        self.position = consumed_positions[index - 1] + 1;
+        // return the name of the local variable before `in` keyword
+        return Ok((TokenType::Name, TokenValue::Name(parts.to_vec().into())));
+      }
+    }
+
     // begin with with the longest name containing all parts
     let mut part_count = parts.len();
     let flattened_keys = self.scope.flatten_keys();
@@ -630,44 +651,31 @@ impl<'lexer> Lexer<'lexer> {
 
     // build the name from name parts
     let name: Name = parts.to_vec().into();
+
     // ------------------------------------------------------------------------
-    // tweak with names of built-in types
+    // tweak with built-in type names
     // ------------------------------------------------------------------------
     if self.type_name
       && matches!(
         name.to_string().as_str(),
-        "boolean" | "number" | "string" | "date" | "date and time" | "time" | "years and months duration" | "days and time duration"
+        "Any" | "Null" | "boolean" | "number" | "string" | "date" | "date and time" | "time" | "years and months duration" | "days and time duration"
       )
     {
-      return Ok((TokenType::Name, TokenValue::Name(name)));
+      self.type_name = false;
+      return Ok((TokenType::BuiltInTypeName, TokenValue::BuiltInTypeName(name)));
     }
+
     // ------------------------------------------------------------------------
     // tweak with date and time literals
     // ------------------------------------------------------------------------
     if matches!(name.to_string().as_str(), "date" | "date and time" | "time" | "duration") {
       return Ok((TokenType::NameDateTime, TokenValue::NameDateTime(name)));
     }
-    // by default return the name as it appears in input
+
+    // by default return the name as it appears on input
     Ok((TokenType::Name, TokenValue::Name(name)))
   }
 
-  /// Takes a list of parts that make up the name and builds the name token.
-  /// If there is one or more dots '.' as parts in the name,
-  /// then this name is a [Token::QualifiedName].
-  /// If there are no dots as parts, then this is a [Token::Name].
-  // fn prepare_name_token(&self, parts: &[String]) -> (TokenType, TokenValue) {
-  //   let mut names: Vec<Name> = vec![];
-  //   for parts in parts.split(|v| v == ".") {
-  //     names.push(parts.to_vec().into());
-  //   }
-  //   if names.len() > 1 {
-  //     (TokenType::QualifiedName, Some(TokenValue::QualifiedName(names)))
-  //   } else {
-  //     (TokenType::Name, Some(TokenValue::Name(parts.to_vec().into())))
-  //   }
-  //   (TokenType::Name, TokenValue::Name(parts.to_vec().into()))
-  // }
-
   /// Consumes a HEX digit from input or reports an error.
   fn consume_hex_digit(&mut self) -> Result<u64> {
     if let Some(ch) = self.char_at(0) {
@@ -875,6 +883,50 @@ impl<'lexer> Lexer<'lexer> {
       false
     }
   }
+
+  /// Returns `true` when the specified character is a function keyword separator,
+  /// that is a character that is allowed after keyword `function`.
+  fn is_function_separator(&self, offset: usize) -> bool {
+    self.is_next_character(&['(', '<'], offset)
+  }
+
+  /// Returns `true` when the specified character is a context keyword separator,
+  /// that is a character that is allowed after keyword `context`.
+  fn is_context_separator(&self, offset: usize) -> bool {
+    self.is_next_character(&['<'], offset)
+  }
+
+  /// Returns `true` when the specified character is a range keyword separator,
+  /// that is a character that is allowed after keyword `range`.
+  fn is_range_separator(&self, offset: usize) -> bool {
+    self.is_next_character(&['<'], offset)
+  }
+
+  /// Returns `true` when the next character on input is a `list` keyword separator,
+  /// that is, a character that is allowed after keyword `list`.
+  /// Allowed characters after keyword `list` are whitespace and `<`.
+  fn is_list_separator(&self, offset: usize) -> bool {
+    self.is_next_character(&['<'], offset)
+  }
+
+  /// Returns `true` when the next character on input is a character
+  /// from specified `chars` list.
+  fn is_next_character(&self, chars: &[char], mut offset: usize) -> bool {
+    while let Some(ch) = self.char_at(offset) {
+      if chars.contains(&ch) {
+        return true;
+      } else if !is_whitespace(ch) {
+        return false;
+      }
+      offset += 1;
+    }
+    false
+  }
+}
+
+/// Returns `true` when the specified character is a separator (white space equivalent).
+fn is_separator(ch: char) -> bool {
+  matches!(ch, WS | '=' | '!' | '<' | '>' | '+' | '-' | '*' | '/' | '%' | '.' | ',' | ')' | '[' | ']' | '}')
 }
 
 /// Returns **true** when the specified character is an ASCII digit.
@@ -887,29 +939,6 @@ fn is_hex_digit(ch: char) -> bool {
   ch.is_digit(16)
 }
 
-/// Returns `true` when the specified character is a separator (white space equivalent).
-fn is_separator(ch: char) -> bool {
-  matches!(ch, WS | '=' | '!' | '<' | '>' | '+' | '-' | '*' | '/' | '%' | '.' | ',' | ')' | '[' | ']' | '}')
-}
-
-/// Returns `true` when the specified character is a context keyword separator,
-/// that is a character that is allowed after keyword `context`.
-fn is_context_separator(ch: char) -> bool {
-  matches!(ch, WS | '<')
-}
-
-/// Returns `true` when the specified character is a function keyword separator,
-/// that is a character that is allowed after keyword `function`.
-fn is_function_separator(ch: char) -> bool {
-  matches!(ch, WS | '(' | '<')
-}
-
-/// Returns `true` when the specified character is a range keyword separator,
-/// that is a character that is allowed after keyword `range`.
-fn is_range_separator(ch: char) -> bool {
-  matches!(ch, WS | '<')
-}
-
 /// Returns `true` when the specified character is a keyword `not` separator,
 /// that is a character that is allowed after keyword `not` at the beginning of unary tests.
 fn is_keyword_not_separator(ch: char) -> bool {
@@ -954,11 +983,6 @@ fn is_vertical_space(ch: char) -> bool {
   matches!(ch, '\u{000A}'..='\u{000D}')
 }
 
-/// Returns `true` when the specified character is a whitespace character or left parenthesis.
-// fn is_whitespace_or_left_paren(ch: char) -> bool {
-//   is_whitespace(ch) || ch == '('
-// }
-
 /// Return the decimal value that corresponds to hexadecimal digit, case insensitive.
 fn hex_to_decimal(ch: char) -> u64 {
   match ch {
@@ -978,7 +1002,7 @@ fn hex_to_decimal(ch: char) -> u64 {
     'D' | 'd' => 13,
     'E' | 'e' => 14,
     'F' | 'f' => 15,
-    _ => 0,
+    _ => u64::MAX,
   }
 }
 
@@ -986,7 +1010,6 @@ fn hex_to_decimal(ch: char) -> u64 {
 /// from additional whitespaces at both sides. The final string is also trimmed.
 /// After trimming, spaces around additional characters (`.`,`/`,`-`,`'`,`+`,`*`) are removed.
 fn flatten_name_parts(parts: &[String]) -> String {
-  //TODO remove this function, this should not be needed, converting Name to string should suffice
   parts
     .iter()
     .map(|s| s.trim().to_string())
@@ -1004,21 +1027,15 @@ fn flatten_name_parts(parts: &[String]) -> String {
 
 /// Definitions of errors raised by the lexer.
 pub mod errors {
-  use dmntk_feel::dmntk_common::DmntkError;
+  use dmntk_common::DmntkError;
 
   /// Lexer errors.
   #[derive(Debug, PartialEq)]
   enum LexerError {
-    // LexerNoMoreTokens,
     UnexpectedEof,
-    // LexerUnexpectedCharacterAtPosition(char, usize),
     ExpectedCharacter(char, char),
     ExpectedCharacters(Vec<char>, char),
     ExpectedHexDigit(char),
-    // NotBuiltInFunctionName(String),
-    // NotBuiltInDateAndTimeFunctionName(String),
-    // NotBuiltInTypeName(String),
-    // NameNotFoundInContext(String),
     UnicodeValueOutOfRange(u64),
     UnicodeSurrogateOutOfRange(u64),
     UnicodeConversionFailed(u64),
@@ -1033,15 +1050,9 @@ pub mod errors {
   impl std::fmt::Display for LexerError {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
       match self {
-        // LexerError::LexerNoMoreTokens => {
-        //   write!(f, "no more tokens on input")
-        // }
         LexerError::UnexpectedEof => {
           write!(f, "unexpected end of file")
         }
-        // LexerError::LexerUnexpectedCharacterAtPosition(ch, pos) => {
-        //   write!(f, "unexpected character: '{}' at position {}", ch, pos)
-        // }
         LexerError::ExpectedCharacter(expected, actual) => {
           write!(f, "expected '{}' character but encountered '{}'", expected, actual)
         }
@@ -1051,18 +1062,6 @@ pub mod errors {
         LexerError::ExpectedHexDigit(actual) => {
           write!(f, "expected hex digit but encountered '{}'", actual)
         }
-        // LexerError::NotBuiltInFunctionName(name) => {
-        //   write!(f, "encountered name is not built-in function name: {}", name)
-        // }
-        // LexerError::NotBuiltInDateAndTimeFunctionName(name) => {
-        //   write!(f, "encountered name is not built-in date and time function name: {}", name)
-        // }
-        // LexerError::NotBuiltInTypeName(name) => {
-        //   write!(f, "encountered name is not built-in type name: {}", name)
-        // }
-        // LexerError::NameNotFoundInContext(name) => {
-        //   write!(f, "encountered name was not found in context: {}", name)
-        // }
         LexerError::UnicodeValueOutOfRange(value) => {
           write!(f, "Unicode value is out of allowed range 0x0000..0x10FFFF : {:X}", value)
         }
@@ -1080,26 +1079,6 @@ pub mod errors {
     LexerError::UnexpectedEof.into()
   }
 
-  // pub fn no_more_tokens() -> DmntkError {
-  //   LexerError::LexerNoMoreTokens.into()
-  // }
-
-  // pub fn name_not_found_in_context(name: &str) -> DmntkError {
-  //   LexerError::NameNotFoundInContext(name.to_string()).into()
-  // }
-
-  // pub fn not_built_in_type_name(name: &str) -> DmntkError {
-  //   LexerError::NotBuiltInTypeName(name.to_string()).into()
-  // }
-
-  // pub fn not_built_in_function_name(name: &str) -> DmntkError {
-  //   LexerError::NotBuiltInFunctionName(name.to_string()).into()
-  // }
-
-  // pub fn not_built_in_date_time_function_name(name: &str) -> DmntkError {
-  //   LexerError::NotBuiltInDateAndTimeFunctionName(name.to_string()).into()
-  // }
-
   pub fn expected_character(expected: char, actual: char) -> DmntkError {
     LexerError::ExpectedCharacter(expected, actual).into()
   }
@@ -1128,6 +1107,7 @@ pub mod errors {
 #[cfg(test)]
 mod tests {
   use super::{flatten_name_parts, is_additional_name_symbol, is_separator};
+  use crate::lexer::hex_to_decimal;
 
   #[test]
   fn test_is_separator() {
@@ -1187,4 +1167,32 @@ mod tests {
     assert_eq!("", flatten_name_parts(&["".to_string(), "".to_string(), "".to_string()]));
     assert_eq!("", flatten_name_parts(&[]));
   }
+
+  #[test]
+  fn test_hex_to_decimal() {
+    assert_eq!(0, hex_to_decimal('0'));
+    assert_eq!(1, hex_to_decimal('1'));
+    assert_eq!(2, hex_to_decimal('2'));
+    assert_eq!(3, hex_to_decimal('3'));
+    assert_eq!(4, hex_to_decimal('4'));
+    assert_eq!(5, hex_to_decimal('5'));
+    assert_eq!(6, hex_to_decimal('6'));
+    assert_eq!(7, hex_to_decimal('7'));
+    assert_eq!(8, hex_to_decimal('8'));
+    assert_eq!(9, hex_to_decimal('9'));
+    assert_eq!(10, hex_to_decimal('A'));
+    assert_eq!(11, hex_to_decimal('B'));
+    assert_eq!(12, hex_to_decimal('C'));
+    assert_eq!(13, hex_to_decimal('D'));
+    assert_eq!(14, hex_to_decimal('E'));
+    assert_eq!(15, hex_to_decimal('F'));
+    assert_eq!(10, hex_to_decimal('a'));
+    assert_eq!(11, hex_to_decimal('b'));
+    assert_eq!(12, hex_to_decimal('c'));
+    assert_eq!(13, hex_to_decimal('d'));
+    assert_eq!(14, hex_to_decimal('e'));
+    assert_eq!(15, hex_to_decimal('f'));
+    assert_eq!(u64::MAX, hex_to_decimal('G'));
+    assert_eq!(u64::MAX, hex_to_decimal('g'));
+  }
 }
